@@ -38,6 +38,7 @@ module Codec.Archive.Zip
 
        -- * Pure functions for working with zip archives
        , toArchive
+       , toArchiveOrFail
        , fromArchive
        , filesInArchive
        , addEntryToArchive
@@ -57,27 +58,21 @@ module Codec.Archive.Zip
 import System.Time ( toUTCTime, addToClockTime, CalendarTime (..), ClockTime (..), TimeDiff (..) )
 #if MIN_VERSION_directory(1,2,0)
 import Data.Time.Clock.POSIX ( utcTimeToPOSIXSeconds )
-#else
 #endif
 import Data.Bits ( shiftL, shiftR, (.&.) )
 import Data.Binary
 import Data.Binary.Get
 import Data.Binary.Put
-import Data.List ( nub, find )
+import Data.List ( nub, find, intercalate )
 import Text.Printf
 import System.FilePath
 import System.Directory ( doesDirectoryExist, getDirectoryContents, createDirectoryIfMissing )
 import Control.Monad ( when, unless, zipWithM )
-#if MIN_VERSION_directory(1,2,0)
-import Control.Monad ( liftM )
-#endif
 import System.Directory ( getModificationTime )
 import System.IO ( stderr, hPutStrLn )
 import qualified Data.Digest.CRC32 as CRC32
 import qualified Data.Map as M
-#if MIN_VERSION_binary(0, 6, 0)
 import Control.Applicative
-#endif
 #ifndef _WINDOWS
 import System.Posix.Files ( setFileTimes )
 #endif
@@ -85,24 +80,13 @@ import System.Posix.Files ( setFileTimes )
 -- from bytestring
 import qualified Data.ByteString.Lazy as B
 
--- from utf8-string
-import Data.ByteString.Lazy.UTF8 ( toString, fromString )
+-- text
+import qualified Data.Text.Lazy as TL
+import qualified Data.Text.Lazy.Encoding as TL
 
 -- from zlib
 import qualified Codec.Compression.Zlib.Raw as Zlib
 
-#if !MIN_VERSION_binary(0, 6, 0)
-manySig :: Word32 -> Get a -> Get [a]
-manySig sig p = do
-    sig' <- lookAhead getWord32le
-    if sig == sig'
-        then do
-            r <- p
-            rs <- manySig sig p
-            return $ r : rs
-        else return []
-#endif
-
 ------------------------------------------------------------------------
 
 -- | Structured representation of a zip archive, including directory
@@ -113,6 +97,10 @@ data Archive = Archive
                 , zComment                :: B.ByteString         -- ^ Comment for whole zip archive
                 } deriving (Read, Show)
 
+instance Binary Archive where
+  put = putArchive
+  get = getArchive
+
 -- | Representation of an archived file, including content and metadata.
 data Entry = Entry
                { eRelativePath            :: FilePath            -- ^ Relative path, using '/' as separator
@@ -136,6 +124,8 @@ data CompressionMethod = Deflate
 -- | Options for 'addFilesToArchive' and 'extractFilesFromArchive'.
 data ZipOption = OptRecursive               -- ^ Recurse into directories when adding files
                | OptVerbose                 -- ^ Print information to stderr
+               | OptDestination FilePath    -- ^ Directory in which to extract
+               | OptLocation FilePath Bool  -- ^ Where to place file when adding files and whether to append current path
                deriving (Read, Show, Eq)
 
 -- | A zip archive with no contents.
@@ -147,11 +137,18 @@ emptyArchive = Archive
 
 -- | Reads an 'Archive' structure from a raw zip archive (in a lazy bytestring).
 toArchive :: B.ByteString -> Archive
-toArchive = runGet getArchive
+toArchive = decode
+
+-- | Like 'toArchive', but returns an 'Either' value instead of raising an error
+-- if the archive cannot be decoded.
+toArchiveOrFail :: B.ByteString -> Either String Archive
+toArchiveOrFail bs = case decodeOrFail bs of
+                           Left (_,_,e)  -> Left e
+                           Right (_,_,x) -> Right x
 
 -- | Writes an 'Archive' structure to a raw zip archive (in a lazy bytestring).
 fromArchive :: Archive -> B.ByteString
-fromArchive = runPut . putArchive
+fromArchive = encode
 
 -- | Returns a list of files in a zip archive.
 filesInArchive :: Archive -> [FilePath]
@@ -167,13 +164,13 @@ addEntryToArchive entry archive =
 -- | Deletes an entry from a zip archive.
 deleteEntryFromArchive :: FilePath -> Archive -> Archive
 deleteEntryFromArchive path archive =
-  let path'      = zipifyFilePath path
-      newEntries = filter (\e -> eRelativePath e /= path') $ zEntries archive
-  in  archive { zEntries = newEntries }
+  archive { zEntries = [e | e <- zEntries archive
+                       , not (eRelativePath e `matches` path)] }
 
 -- | Returns Just the zip entry with the specified path, or Nothing.
 findEntryByPath :: FilePath -> Archive -> Maybe Entry
-findEntryByPath path archive = find (\e -> path == eRelativePath e) (zEntries archive)
+findEntryByPath path archive =
+  find (\e -> path `matches` eRelativePath e) (zEntries archive)
 
 -- | Returns uncompressed contents of zip entry.
 fromEntry :: Entry -> B.ByteString
@@ -198,7 +195,7 @@ toEntry path modtime contents =
            then (NoCompression, contents, uncompressedSize)
            else (Deflate, compressedData, compressedSize)
       crc32 = CRC32.crc32 contents
-  in  Entry { eRelativePath            = path
+  in  Entry { eRelativePath            = normalizePath path
             , eCompressionMethod       = compressionMethod
             , eLastModified            = modtime
             , eCRC32                   = crc32
@@ -215,13 +212,19 @@ toEntry path modtime contents =
 readEntry :: [ZipOption] -> FilePath -> IO Entry
 readEntry opts path = do
   isDir <- doesDirectoryExist path
-  let path' = zipifyFilePath $ normalise $
-              path ++ if isDir then "/" else ""  -- make sure directories end with /
+  -- make sure directories end in / and deal with the OptLocation option
+  let path' = let p = path ++ (case reverse path of
+                                    ('/':_) -> ""
+                                    _ | isDir -> "/"
+                                      | otherwise -> "") in
+              (case [(l,a) | OptLocation l a <- opts] of
+                    ((l,a):_) -> if a then l </> p else l
+                    _         -> p)
   contents <- if isDir
                  then return B.empty
                  else B.readFile path
 #if MIN_VERSION_directory(1,2,0)
-  modEpochTime <- liftM (floor . utcTimeToPOSIXSeconds)
+  modEpochTime <- fmap (floor . utcTimeToPOSIXSeconds)
                    $ getModificationTime path
 #else
   (TOD modEpochTime _) <- getModificationTime path
@@ -239,7 +242,9 @@ readEntry opts path = do
 -- | Writes contents of an 'Entry' to a file.
 writeEntry :: [ZipOption] -> Entry -> IO ()
 writeEntry opts entry = do
-  let path = eRelativePath entry
+  let path = case [d | OptDestination d <- opts] of
+                  (x:_) -> x </> eRelativePath entry
+                  _     -> eRelativePath entry
   -- create directories if needed
   let dir = takeDirectory path
   exists <- doesDirectoryExist dir
@@ -275,20 +280,25 @@ addFilesToArchive opts archive files = do
 -- Note that the last-modified time is set correctly only in POSIX,
 -- not in Windows.
 extractFilesFromArchive :: [ZipOption] -> Archive -> IO ()
-extractFilesFromArchive opts archive = mapM_ (writeEntry opts) $ zEntries archive
+extractFilesFromArchive opts archive =
+  mapM_ (writeEntry opts) $ zEntries archive
 
 --------------------------------------------------------------------------------
 -- Internal functions for reading and writing zip binary format.
 
 -- Note that even on Windows, zip files use "/" internally as path separator.
-zipifyFilePath :: FilePath -> String
-zipifyFilePath path =
-  let dir = takeDirectory path
-      fn  = takeFileName path
+normalizePath :: FilePath -> String
+normalizePath path =
+  let dir   = takeDirectory path
+      fn    = takeFileName path
       (_drive, dir') = splitDrive dir
       -- note: some versions of filepath return ["."] if no dir
-      dirParts = dropWhile (==".") $ splitDirectories dir'
-  in  (concat (map (++ "/") dirParts)) ++ fn
+      dirParts = filter (/=".") $ splitDirectories dir'
+  in  intercalate "/" (dirParts ++ [fn])
+
+-- Equality modulo normalization.  So, "./foo" `matches` "foo".
+matches :: FilePath -> FilePath -> Bool
+matches fp1 fp2 = normalizePath fp1 == normalizePath fp2
 
 -- | Uncompress a lazy bytestring.
 compressData :: CompressionMethod -> B.ByteString -> B.ByteString
@@ -425,15 +435,9 @@ setFileTimeStamp file epochtime = do
 
 getArchive :: Get Archive
 getArchive = do
-#if MIN_VERSION_binary(0, 6, 0)
   locals <- many getLocalFile
   files <- many (getFileHeader (M.fromList locals))
   digSig <- Just `fmap` getDigitalSignature <|> return Nothing
-#else
-  locals <- manySig 0x04034b50 getLocalFile
-  files <- manySig 0x02014b50 (getFileHeader (M.fromList locals))
-  digSig <- lookAheadM getDigitalSignature
-#endif
   endSig <- getWord32le
   unless (endSig == 0x06054b50)
     $ fail "Did not find end of central directory signature"
@@ -473,13 +477,13 @@ putArchive archive = do
 fileHeaderSize :: Entry -> Word32
 fileHeaderSize f =
   fromIntegral $ 4 + 2 + 2 + 2 + 2 + 2 + 2 + 4 + 4 + 4 + 2 + 2 + 2 + 2 + 2 + 4 + 4 +
-    fromIntegral (B.length $ fromString $ zipifyFilePath $ eRelativePath f) +
+    fromIntegral (B.length $ fromString $ normalizePath $ eRelativePath f) +
     B.length (eExtraField f) + B.length (eFileComment f)
 
 localFileSize :: Entry -> Word32
 localFileSize f =
   fromIntegral $ 4 + 2 + 2 + 2 + 2 + 2 + 4 + 4 + 4 + 2 + 2 +
-    fromIntegral (B.length $ fromString $ zipifyFilePath $ eRelativePath f) +
+    fromIntegral (B.length $ fromString $ normalizePath $ eRelativePath f) +
     B.length (eExtraField f) + B.length (eCompressedData f)
 
 -- Local file header:
@@ -540,27 +544,27 @@ getLocalFile = do
   return (fromIntegral offset, compressedData)
 
 getWordsTilSig :: Word32 -> Get [Word8]
-getWordsTilSig sig = do
+getWordsTilSig sig = go []
+  where
+    go acc = do
 #if MIN_VERSION_binary(0, 6, 0)
-  (getWord32le >>= ensure (== sig) >> return []) <|>
-    do w <- getWord8
-       ws <- getWordsTilSig sig
-       return (w:ws)
+      (getWord32le >>= ensure (== sig) >> return (reverse acc)) <|>
+        do w <- getWord8
+           go (w:acc)
 #else
-    sig' <- lookAhead getWord32le
-    if sig == sig'
-        then skip 4 >> return []
-        else do
-            w <- getWord8
-            ws <- getWordsTilSig sig
-            return (w:ws)
+      sig' <- lookAhead getWord32le
+      if sig == sig'
+          then skip 4 >> return (reverse acc)
+          else do
+              w <- getWord8
+              go (w:acc)
 #endif
 
 putLocalFile :: Entry -> Put
 putLocalFile f = do
   putWord32le 0x04034b50
   putWord16le 20 -- version needed to extract (>=2.0)
-  putWord16le 2  -- general purpose bit flag (max compression)
+  putWord16le 0x802  -- general purpose bit flag (bit 1 = max compression, bit 11 = UTF-8)
   putWord16le $ case eCompressionMethod f of
                      NoCompression -> 0
                      Deflate       -> 8
@@ -571,9 +575,9 @@ putLocalFile f = do
   putWord32le $ eCompressedSize f
   putWord32le $ eUncompressedSize f
   putWord16le $ fromIntegral $ B.length $ fromString
-              $ zipifyFilePath $ eRelativePath f
+              $ normalizePath $ eRelativePath f
   putWord16le $ fromIntegral $ B.length $ eExtraField f
-  putLazyByteString $ fromString $ zipifyFilePath $ eRelativePath f
+  putLazyByteString $ fromString $ normalizePath $ eRelativePath f
   putLazyByteString $ eExtraField f
   putLazyByteString $ eCompressedData f
 
@@ -658,7 +662,7 @@ putFileHeader offset local = do
   putWord32le 0x02014b50
   putWord16le 0  -- version made by
   putWord16le 20 -- version needed to extract (>= 2.0)
-  putWord16le 2  -- general purpose bit flag (max compression)
+  putWord16le 0x802  -- general purpose bit flag (bit 1 = max compression, bit 11 = UTF-8)
   putWord16le $ case eCompressionMethod local of
                      NoCompression -> 0
                      Deflate       -> 8
@@ -669,14 +673,14 @@ putFileHeader offset local = do
   putWord32le $ eCompressedSize local
   putWord32le $ eUncompressedSize local
   putWord16le $ fromIntegral $ B.length $ fromString
-              $ zipifyFilePath $ eRelativePath local
+              $ normalizePath $ eRelativePath local
   putWord16le $ fromIntegral $ B.length $ eExtraField local
   putWord16le $ fromIntegral $ B.length $ eFileComment local
   putWord16le 0  -- disk number start
   putWord16le $ eInternalFileAttributes local
   putWord32le $ eExternalFileAttributes local
   putWord32le offset
-  putLazyByteString $ fromString $ zipifyFilePath $ eRelativePath local
+  putLazyByteString $ fromString $ normalizePath $ eRelativePath local
   putLazyByteString $ eExtraField local
   putLazyByteString $ eFileComment local
 
@@ -686,22 +690,11 @@ putFileHeader offset local = do
 -- >     size of data                    2 bytes
 -- >     signature data (variable size)
 
-#if MIN_VERSION_binary(0, 6, 0)
 getDigitalSignature :: Get B.ByteString
 getDigitalSignature = do
   getWord32le >>= ensure (== 0x05054b50)
   sigSize <- getWord16le
   getLazyByteString (toEnum $ fromEnum sigSize)
-#else
-getDigitalSignature :: Get (Maybe B.ByteString)
-getDigitalSignature = do
-  hdrSig <- getWord32le
-  if hdrSig /= 0x05054b50
-     then return Nothing
-     else do
-        sigSize <- getWord16le
-        getLazyByteString (toEnum $ fromEnum sigSize) >>= return . Just
-#endif
 
 putDigitalSignature :: Maybe B.ByteString -> Put
 putDigitalSignature Nothing = return ()
@@ -715,3 +708,9 @@ ensure p val =
   if p val
      then return ()
      else fail "ensure not satisfied"
+
+toString :: B.ByteString -> String
+toString = TL.unpack . TL.decodeUtf8
+
+fromString :: String -> B.ByteString
+fromString = TL.encodeUtf8 . TL.pack
